#ifndef FGGL_DATA_MODEL_H
#define FGGL_DATA_MODEL_H

#include <tuple>
#include <vector>
#include <fggl/math/types.hpp>

namespace fggl::data {

	struct Vertex {
		math::vec3 posititon;
		math::vec3 normal;
		math::vec3 colour;
	};

	// comparison operators

	inline bool operator<(const Vertex& lhs, const Vertex& rhs) {
		return std::tie( lhs.posititon, lhs.normal, lhs.colour )
		     < std::tie( rhs.posititon, rhs.normal, rhs.colour );
	}

	inline bool operator==(const Vertex& lhs, const Vertex& rhs) {
		return lhs.posititon == rhs.posititon
			&& lhs.colour == rhs.colour
			&& lhs.normal == rhs.normal;
	}

	inline bool operator!=(const Vertex& lhs, const Vertex& rhs) {
		return !(lhs==rhs);
	}

	class Mesh {
		public:
			Mesh();
			~Mesh() = default;

			/**
			 * Push a vertex with deduplication.
			 *
			 * If the vertex is already known, the old vertex will be used and a new entry will
			 * be added to the indicies list.
			 */
			inline void push(Vertex vert) {
				auto idx = indexOf(vert);
				if ( idx == -1 ) {
                    idx = pushVertex(vert);
                    pushIndex(idx);
                } else {
                    pushIndex(idx);
                }
			}

			/**
			 * Push a new vertex index.
			 *
			 * Don't push non-existant indexes, it'll make the rendering backend sad.
			 */
			void pushIndex(unsigned int idx);

			/**
			 * Force push a vertex duplicate.
			 *
			 * If the vetex already exists, it will be duplicated.
			 */
			unsigned int pushVertex(Vertex vert);

			/**
			 * remove Duplicate verticies
			 *
			 * If generated using the procedural system, the mesh data can get a little messy.
			 * This method will remove unused and duplicate verticies.
			 */
			void removeDups();

			/**
			 * Search for a vertex.
			 *
			 * Returns the index of a vert, else returns -1.
			 */
			int indexOf(Vertex vert);

			inline std::vector<Vertex>& vertexList() {
				return m_verts;
			}

			inline std::size_t vertexCount() {
				return m_verts.size();
			}

			inline std::vector<uint32_t>& indexList() {
				return m_index;
			}

			inline std::size_t indexCount() {
				return m_index.size();
			}

			inline Vertex& vertex(int idx) {
				return m_verts[idx];
			}
            unsigned int restartVertex;

		private:
			std::vector<Vertex> m_verts;
			std::vector<unsigned int> m_index;
	};

	class Model {
		public:
			Model() = default;
			~Model() = default;

			inline void append(const Mesh& mesh) {
				m_meshes.push_back( mesh );
			}

		private:
			std::vector<Mesh> m_meshes;
	};

}

#endif
